using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Microsoft.Xna.Framework;

namespace Frets.Game {
	/// <summary>
	/// Represents the basic functionality of the game.
	/// </summary>
	partial class Game {

		#region Properties

		private readonly string BaseDirectory;

		/// <summary>
		/// Gets the base directory for the game data.
		/// </summary>
		public string DataDirectory {
			get { return Path.Combine(this.BaseDirectory, "data"); }
		}

		/// <summary>
		/// Gets the base directory for songs.
		/// </summary>
		public string SongDirectory {
			get { return Path.Combine(this.DataDirectory, "songs"); }
		}

		private readonly SoundEffects effects;
		/// <summary>
		/// Gets the game's sound effects class.
		/// </summary>
		public SoundEffects SoundEffects {
			get { return this.effects; }
		}

		#endregion


		/// <summary>
		/// Current song.
		/// </summary>
		private Song currentSong;
		public Song CurrentSong {
			get { return this.currentSong; }
		}

		/// <summary>
		/// Sound system to play on.
		/// </summary>
		private readonly FmodEx.Fmod SoundSystem;

		public Game(FmodEx.Fmod soundSystem, string baseDirectory) {
			this.SoundSystem = soundSystem;
			this.BaseDirectory = baseDirectory;
			this.effects = new SoundEffects(this.SoundSystem, this.DataDirectory);
		}


		public void LoadSong(Song song) {
			this.currentSong = song;
			this.currentSong.LoadSound();
		}
		public void LoadSong(string name) {
			this.currentSong = new Song(this.SoundSystem, Path.Combine(this.SongDirectory, name));
			this.currentSong.LoadSound();
		}

		public void StartSong(Song.Notes.DifficultyModes difficulty) {
			this.CurrentSong.Start(difficulty);
		}


		private bool Pick0WasHeld = false;
		private bool Pick1WasHeld = false;


		private int CurrentFretEvent = 0;
		private bool AdvanceNotePending = false;

		private Song.Notes.HeldFrets ExpectedFrets = Song.Notes.HeldFrets.None;
		private bool HitLastNoteSuccessfully = false;


		public void Update() {
			if (this.CurrentSong != null && this.CurrentFretEvent < this.CurrentSong.FretEvents.Length) {

				bool ChangeGuitarPlaying = this.CurrentSong.GuitarPlaying;

				TimeSpan CurrentTime = this.CurrentSong.Position;
				bool AdvancedANote = false;
				bool HavePluckedSuccesfully = false;

				while (CurrentFretEvent < this.CurrentSong.FretEvents.Length && this.CurrentSong.FretEvents[CurrentFretEvent].EndTime <= CurrentTime) {
					++CurrentFretEvent;
					AdvanceNotePending = true;
				}

				if (CurrentFretEvent < this.CurrentSong.FretEvents.Length) {

					if (AdvanceNotePending && this.CurrentSong.FretEvents[CurrentFretEvent].StartTime <= CurrentTime) {
						AdvancedANote = true;
						HitLastNoteSuccessfully = false;
					}


					if (this.Pick0Pressed && !this.Pick0WasHeld) {
						// We have plucked!
						Frets.Game.Song.Notes.HeldFrets FretsHeld = CurrentHeldFrets;

						for (int i = 0; i < this.CurrentSong.FretEvents.Length; ++i) {
							if (Math.Abs(((TimeSpan)(CurrentSong.FretEvents[i].StartTime - CurrentTime)).TotalMilliseconds) < 100d) {
								if (CurrentSong.FretEvents[i].Frets == this.CurrentHeldFrets) {
									CurrentSong.FretEvents[i].NoteHit = true;
									ExpectedFrets = CurrentSong.FretEvents[i].Frets;
								}
								HavePluckedSuccesfully = true;
								break;
							}
						}
						if (!HavePluckedSuccesfully) {
							this.SoundEffects.Mistake.Play();
						}
					}

					if (AdvancedANote || HavePluckedSuccesfully) ChangeGuitarPlaying = (CurrentSong.FretEvents[CurrentFretEvent].NoteHit);

					this.Pick0WasHeld = this.Pick0Pressed;
					this.Pick1WasHeld = this.Pick1Pressed;

					this.CurrentSong.GuitarPlaying = ChangeGuitarPlaying;
					HitLastNoteSuccessfully = ChangeGuitarPlaying;

				}

			}
		}

		#region Frets / Picks

		/// <summary>
		/// Gets or sets whether the first fret is held.
		/// </summary>
		private bool fret0Pressed;
		public bool Fret0Pressed {
			get { return this.fret0Pressed; }
			set { this.fret0Pressed = value; }
		}
		private bool fret1Pressed;
		/// <summary>
		/// Gets or sets whether the second fret is held.
		/// </summary>
		public bool Fret1Pressed {
			get { return this.fret1Pressed; }
			set { this.fret1Pressed = value; }
		}
		/// <summary>
		/// Gets or sets whether the third fret is held.
		/// </summary>
		private bool fret2Pressed;
		public bool Fret2Pressed {
			get { return this.fret2Pressed; }
			set { this.fret2Pressed = value; }
		}
		/// <summary>
		/// Gets or sets whether the fourth fret is held.
		/// </summary>
		private bool fret3Pressed;
		public bool Fret3Pressed {
			get { return this.fret3Pressed; }
			set { this.fret3Pressed = value; }
		}
		/// <summary>
		/// Gets or sets whether the fifth fret is held.
		/// </summary>
		private bool fret4Pressed;
		public bool Fret4Pressed {
			get { return this.fret4Pressed; }
			set { this.fret4Pressed = value; }
		}

		/// <summary>
		/// Gets or sets whether the pick is pressed in first direction.
		/// </summary>
		private bool pick0Pressed;
		public bool Pick0Pressed {
			get { return this.pick0Pressed; }
			set { this.pick0Pressed = value; }
		}
		/// <summary>
		/// Gets or sets whether the pick is pressed in the second direction.
		/// </summary>
		private bool pick1Pressed;
		public bool Pick1Pressed {
			get { return this.pick1Pressed; }
			set { this.pick1Pressed = value; }
		}

		private Song.Notes.HeldFrets CurrentHeldFrets {
			get {
				return (this.Fret0Pressed ? Song.Notes.HeldFrets.Fret0 : Song.Notes.HeldFrets.None)
					 | (this.Fret1Pressed ? Song.Notes.HeldFrets.Fret1 : Song.Notes.HeldFrets.None)
					 | (this.Fret2Pressed ? Song.Notes.HeldFrets.Fret2 : Song.Notes.HeldFrets.None)
					 | (this.Fret3Pressed ? Song.Notes.HeldFrets.Fret3 : Song.Notes.HeldFrets.None)
					 | (this.Fret4Pressed ? Song.Notes.HeldFrets.Fret4 : Song.Notes.HeldFrets.None);
			}
		}

		#endregion

		/// <summary>
		/// Get an array of songs from a particular directory.
		/// </summary>
		/// <param name="directory"></param>
		/// <returns></returns>
		public Song[] GetSongs(string directory) {
			List<Song> Songs = new List<Song>();

			string[] RequiredFiles = new string[] { "guitar.ogg", "song.ini", "notes.mid" };


			foreach (string SongDir in Directory.GetDirectories(Path.Combine(this.SongDirectory, directory))) {
				bool IsValid = true;
				foreach (string CheckRequired in RequiredFiles) {
					if (!File.Exists(Path.Combine(SongDir, CheckRequired))) {
						IsValid = false;
						break;
					}
				}
				if (IsValid) {
					Songs.Add(new Song(this.SoundSystem, SongDir));
				}
			}

			Songs.Sort(delegate(Song a, Song b) { return a.Name.CompareTo(b.Name); });

			return Songs.ToArray();
		}

		public string[] GetDirectories(string directory) {
			List<string> Directories = new List<string>();
			foreach (string SongCollectionDir in Directory.GetDirectories(Path.Combine(this.SongDirectory, directory))) {
				if (Directory.GetDirectories(SongCollectionDir).Length > 0) Directories.Add(SongCollectionDir);
			}
			return Directories.ToArray();
		}
	}
}
