using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Frets.Game {
	partial class Song {
		public class Notes {


			private readonly Parsers.StandardMidiFile MidiFile;
			private readonly Parsers.StandardMidiFile.Track MergedTrack;

			/// <summary>
			/// MIDI note values that correspond to a particular fret.
			/// </summary>
			public enum MidiFretNote {
				// "Supaeasy" difficulty.
				SupaeasyFret0 = 60,
				SupaeasyFret1 = 61,
				SupaeasyFret2 = 62,
				SupaeasyFret3 = 63,
				SupaeasyFret4 = 64,
				// "Easy" difficulty.
				EasyFret0 = 72,
				EasyFret1 = 73,
				EasyFret2 = 74,
				EasyFret3 = 75,
				EasyFret4 = 76,
				// "Medium" difficulty.
				MediumFret0 = 84,
				MediumFret1 = 85,
				MediumFret2 = 86,
				MediumFret3 = 87,
				MediumFret4 = 88,
				// "Amazing" difficulty.
				AmazingFret0 = 96,
				AmazingFret1 = 97,
				AmazingFret2 = 98,
				AmazingFret3 = 99,
				AmazingFret4 = 100,
			}

			/// <summary>
			/// Flags corresponding to a particular difficulty mode.
			/// </summary>
			[Flags()] public enum DifficultyModes {
				None = 0,
				Supaeasy = 1,
				Easy = 2,
				Medium = 4,
				Amazing = 8,
				All = Supaeasy | Easy | Medium | Amazing,
			}

			/// <summary>
			/// Flags corresponding to which frets are held.
			/// </summary>
			[Flags()] public enum HeldFrets {
				None = 0,
				Fret0 = 1,
				Fret1 = 2,
				Fret2 = 4,
				Fret3 = 8,
				Fret4 = 16,
			}

			/// <summary>
			/// Convert a MIDI note number to a fret number (ignoring difficulty).
			/// </summary>
			/// <param name="note">The note to convert.</param>
			/// <returns>The fret number corresponding to that note, or None if the note doesn't match a fret.</returns>
			public HeldFrets NoteToFret(MidiFretNote note) {
				switch (note) {
					case MidiFretNote.SupaeasyFret0:
					case MidiFretNote.EasyFret0:
					case MidiFretNote.MediumFret0:
					case MidiFretNote.AmazingFret0:
						return HeldFrets.Fret0;
					case MidiFretNote.SupaeasyFret1:
					case MidiFretNote.EasyFret1:
					case MidiFretNote.MediumFret1:
					case MidiFretNote.AmazingFret1:
						return HeldFrets.Fret1;
					case MidiFretNote.SupaeasyFret2:
					case MidiFretNote.EasyFret2:
					case MidiFretNote.MediumFret2:
					case MidiFretNote.AmazingFret2:
						return HeldFrets.Fret2;
					case MidiFretNote.SupaeasyFret3:
					case MidiFretNote.EasyFret3:
					case MidiFretNote.MediumFret3:
					case MidiFretNote.AmazingFret3:
						return HeldFrets.Fret3;
					case MidiFretNote.SupaeasyFret4:
					case MidiFretNote.EasyFret4:
					case MidiFretNote.MediumFret4:
					case MidiFretNote.AmazingFret4:
						return HeldFrets.Fret4;
					default:
						return HeldFrets.None;
				}
			}

			/// <summary>
			/// Check whether a particular note is valid for a particular difficulty level.
			/// </summary>
			/// <param name="note">The note to check.</param>
			/// <param name="difficulty">The difficulty level.</param>
			/// <returns>True if the note is valid, false otherwise.</returns>
			public bool NoteIsValidForDifficulty(MidiFretNote note, DifficultyModes difficulty) {
				switch (difficulty) {
					case DifficultyModes.Supaeasy:
						return note >= MidiFretNote.SupaeasyFret0 && note <= MidiFretNote.SupaeasyFret4;
					case DifficultyModes.Easy:
						return note >= MidiFretNote.EasyFret0 && note <= MidiFretNote.EasyFret4;
					case DifficultyModes.Medium:
						return note >= MidiFretNote.MediumFret0 && note <= MidiFretNote.MediumFret4;
					case DifficultyModes.Amazing:
						return note >= MidiFretNote.AmazingFret0 && note <= MidiFretNote.AmazingFret4;
					default:
						throw new ArgumentException("Difficulty must be a single difficulty level.");
				}
			}

			/// <summary>
			/// Gets the difficulty modes that are available for this song.
			/// </summary>
			public DifficultyModes AvailableModes {
				get {
					DifficultyModes Result = DifficultyModes.None;
					foreach (Parsers.StandardMidiFile.Track.TrackEvent TE in this.MergedTrack.Events) {
						if (TE.Command.Type == Parsers.StandardMidiFile.Track.TrackEvent.MidiCommand.CommandType.NoteOn) {
							int NoteValue = TE.Command.Data[0];
							if (NoteValue >= (int)MidiFretNote.SupaeasyFret0 && NoteValue <= (int)MidiFretNote.SupaeasyFret4) {
								Result |= DifficultyModes.Supaeasy;
							} else if (NoteValue >= (int)MidiFretNote.EasyFret0 && NoteValue <= (int)MidiFretNote.EasyFret4) {
								Result |= DifficultyModes.Easy;
							} else if (NoteValue >= (int)MidiFretNote.MediumFret0 && NoteValue <= (int)MidiFretNote.MediumFret4) {
								Result |= DifficultyModes.Medium;
							} if (NoteValue >= (int)MidiFretNote.AmazingFret0 && NoteValue <= (int)MidiFretNote.AmazingFret4) {
								Result |= DifficultyModes.Amazing;
							}
						}
						if (Result == DifficultyModes.All) break;
					}
					return Result;
				}
			}


			public Notes(string midiFilename) {

				// Open the MIDI file.
				using (Stream s = File.OpenRead(midiFilename)) {
					this.MidiFile = new Parsers.StandardMidiFile(s);
				}

				this.MergedTrack = this.MidiFile.GetMergedTrack();

			}

			public struct FretEvent {
				public HeldFrets Frets;
				public TimeSpan StartTime;
				public TimeSpan Duration;
				public TimeSpan EndTime {
					get { return this.StartTime + this.Duration; }
				}
				public float NoteSpeed;
				public bool NoteHit;
			}

			public FretEvent[] GetSongFretEvents(DifficultyModes difficulty) {
				List<FretEvent> Events = new List<FretEvent>();
				int? CurrentTempo = null;

				TimeSpan CurrentTime = TimeSpan.Zero;
				float CurrentSpeed = 20f;
				
				foreach (Parsers.StandardMidiFile.Track.TrackEvent TE in this.MergedTrack.Events) {

					// Adjust the current time.
					TimeSpan Delta = TE.ConvertToTimeSpan(TE.DeltaTime, this.MidiFile.DeltaTimeTicksPerQuarter, CurrentTempo);
					CurrentTime = CurrentTime + Delta;


					switch (TE.Command.Type) {
						case Parsers.StandardMidiFile.Track.TrackEvent.MidiCommand.CommandType.MetaEvent:
							switch (TE.Command.MetaType) {
								case Parsers.StandardMidiFile.Track.TrackEvent.MidiCommand.MetaCommandType.SetTempo:
									CurrentTempo = TE.Command.Data[2] + (TE.Command.Data[1] << 8) + (TE.Command.Data[0] << 16);
									CurrentSpeed = 18000000f / (float)CurrentTempo;
									break;
							}
							break;
						case Parsers.StandardMidiFile.Track.TrackEvent.MidiCommand.CommandType.NoteOn:
						case Parsers.StandardMidiFile.Track.TrackEvent.MidiCommand.CommandType.NoteOff:

							MidiFretNote ActiveNote = (MidiFretNote)TE.Command.Data[0];

							if (TE.Command.Type == Frets.Parsers.StandardMidiFile.Track.TrackEvent.MidiCommand.CommandType.NoteOff || TE.Command.Data[1] == 0) {
								if (Events.Count != 0) {
									FretEvent F = Events[Events.Count - 1];
									F.Duration = CurrentTime - F.StartTime;
									Events[Events.Count - 1] = F;
								}
							} else {
								if (this.NoteIsValidForDifficulty(ActiveNote, difficulty)) {
									if (Events.Count != 0 && Events[Events.Count - 1].StartTime == CurrentTime) {
										FretEvent F = Events[Events.Count - 1];
										F.Frets |= this.NoteToFret(ActiveNote);
										Events[Events.Count - 1] = F;
									} else {
										FretEvent F = new FretEvent();
										F.StartTime = CurrentTime;
										F.Frets |= this.NoteToFret(ActiveNote);
										F.NoteSpeed = CurrentSpeed;
										Events.Add(F);
									}
								}
							}
							break;
					}

				}

				return Events.ToArray();
			}

		}
	}
}

