using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace Frets.Interface {
	class Menus {

		/// <summary>
		/// Sound effects class to play menu sounds.
		/// </summary>
		private readonly Game.Game FretsGame;

		/// <summary>
		/// 
		/// </summary>
		/// <param name="soundEffects"></param>
		public Menus(Game.Game fretsGame) {
			this.FretsGame = fretsGame;
		}

		SpriteBatch ForegroundBatch;
		SpriteFont MainFont;

		public class MenuItem {

			private MenuItem lastSelectedItem;
			internal MenuItem LastSelectedItem {
				get { return this.lastSelectedItem ?? (this.ChildMenus.Count > 0 ? this.ChildMenus[0] : null); }
				set { this.lastSelectedItem = value; }
			}

			private string text;
			/// <summary>
			/// Gets or sets the text of the menu.
			/// </summary>
			public string Text {
				get { return this.text; }
				set { this.text = value; }
			}
			private List<MenuItem> childMenus;
			/// <summary>
			/// Gets a list of child menu items.
			/// </summary>
			public List<MenuItem> ChildMenus {
				get { return this.childMenus; }
			}

			private object tag;
			/// <summary>
			/// Gets or sets custom data that can be attached to a menu item.
			/// </summary>
			public object Tag {
				get { return this.tag; }
				set { this.tag = value; }
			}

			private MenuItem parent;
			/// <summary>
			/// Gets or sets this menu's parent.
			/// </summary>
			public MenuItem Parent {
				get { return this.parent;}
				set { this.parent = value;}
			}

			public MenuItem(MenuItem parent, string text, object tag) {
				this.parent = parent;
				this.text = text;
				this.tag = tag;
				this.childMenus = new List<MenuItem>();
			}

			public MenuItem(MenuItem parent, string text)
				: this(parent, text, null) {
			}
			public MenuItem(MenuItem parent)
				: this(parent, null) {
			}

			public event EventHandler Entered;
			internal virtual void OnEntered(EventArgs e) {
				if (Entered != null) Entered(this, e);
			}


		}

		private MenuItem currentMenu;
		/// <summary>
		/// The current menu root to display.
		/// </summary>
		public MenuItem CurrentMenu {
			get { return this.currentMenu; }
			set { this.currentMenu = value; }
		}

		private MenuItem selectedMenu;
		/// <summary>
		/// The currently selected menu.
		/// </summary>
		public MenuItem SelectedMenu {
			get { return this.selectedMenu; }
			set { this.selectedMenu = value; }
		}

		public void LoadContent(ContentManager content, GraphicsDeviceManager graphics) {
			this.ForegroundBatch = new SpriteBatch(graphics.GraphicsDevice);
			this.MainFont = content.Load<SpriteFont>("Content/Fonts/MainFont");
		}


		/// <summary>
		/// Render the menus.
		/// </summary>
		public void Draw(GraphicsDeviceManager graphics) {

			graphics.GraphicsDevice.Clear(Color.Black);

			if (this.CurrentMenu == null) return;

			List<string> StringsToDraw = new List<string>();
			float MenuHeights = 0;
			foreach (MenuItem MI in CurrentMenu.ChildMenus) {
				StringsToDraw.Add(MI.Text);
				MenuHeights += this.MainFont.MeasureString(MI.Text).Y;
			}
			MenuHeights *= 0.5f;

			ForegroundBatch.Begin();

			float CurrentY = (graphics.GraphicsDevice.Viewport.Height - MenuHeights) / 2f;


			foreach (MenuItem MI in CurrentMenu.ChildMenus) {
				Vector2 StringSize = this.MainFont.MeasureString(MI.Text);
				ForegroundBatch.DrawString(this.MainFont, MI.Text, new Vector2(graphics.GraphicsDevice.Viewport.Width / 2f, CurrentY), MI == this.SelectedMenu ? Color.White : Color.Gray, 0.0f, StringSize / 2f, 0.5f, SpriteEffects.None, 0.5f);
				CurrentY += StringSize.Y * 0.5f;
			}

			
			ForegroundBatch.End();

		}

		/// <summary>
		/// Move to the previous menu item.
		/// </summary>
		public void MoveUp() {
			this.Move(-1);
		}

		/// <summary>
		/// Move to the next menu item.
		/// </summary>
		public void MoveDown() {
			this.Move(+1);
		}

		/// <summary>
		/// Move up or down menu items.
		/// </summary>
		/// <param name="direction">How far to go.</param>
		/// <remarks>direction should only be between -1 and +1.</remarks>
		private void Move(int direction) {
			this.SelectedMenu = this.CurrentMenu.ChildMenus[((this.CurrentMenu.ChildMenus.IndexOf(this.SelectedMenu) + direction) + this.CurrentMenu.ChildMenus.Count) % this.CurrentMenu.ChildMenus.Count];
			this.CurrentMenu.LastSelectedItem = this.SelectedMenu;
			this.FretsGame.SoundEffects.Crunch.Play();
		}

		/// <summary>
		/// Enter a menu item.
		/// </summary>
		public void Enter() {
			this.CurrentMenu = this.SelectedMenu;
			this.CurrentMenu.OnEntered(new EventArgs());

			if (this.CurrentMenu.ChildMenus.Count > 0) {
				this.FretsGame.SoundEffects.MenuIn.Play();
			} else {
				this.CurrentMenu = this.CurrentMenu.Parent;
				this.FretsGame.SoundEffects.Mistake.Play();
			}

			this.SelectedMenu = this.CurrentMenu.LastSelectedItem;
		}

		/// <summary>
		/// Move back out of a menu to its parent.
		/// </summary>
		public void GoBack() {
			if (this.CurrentMenu.Parent == null) {
				this.FretsGame.SoundEffects.Mistake.Play();
			} else {
				this.CurrentMenu = this.CurrentMenu.Parent;
				this.FretsGame.SoundEffects.MenuOut.Play();
			}
			this.SelectedMenu = this.CurrentMenu.LastSelectedItem;
		}


	}
}
