using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Frets.Parsers {
	class StandardMidiFile {

		public struct Track {

			/// <summary>
			/// Try and get the name of the track.
			/// </summary>
			/// <param name="name">Returns the name of the track if available.</param>
			/// <returns>True if the track is named, false otherwise.</returns>
			public bool TryGetName(out string name) {
				foreach (TrackEvent T in this.Events) {
					if (T.Command.Type == TrackEvent.MidiCommand.CommandType.MetaEvent && T.Command.MetaType == TrackEvent.MidiCommand.MetaCommandType.TrackName) {
						name = T.Command.DataToString;
						return true;
					}
				}
				name = default(string);
				return false;
			}

            public struct TrackEvent {


				public override string ToString() {
					return "{" + this.DeltaTime.ToString() + " : " + this.Command.ToString() + "}";
				}

                public struct MidiCommand {

					public override string ToString() {
						return "{" + this.Channel.ToString() + ", " + this.Type.ToString() + (this.Type == CommandType.MetaEvent ? " - " + this.MetaType.ToString() : "") + "}";
					}
                    
                    public enum CommandType {
                        NoteOff = 0x80,
                        NoteOn = 0x90,
                        KeyAfterTouch = 0xA0,
                        ControlChange = 0xB0,
                        PatchChange = 0xC0,
                        ChannelAfterTouch = 0xD0,
                        PitchWheelChange = 0xE0,
                        MetaEvent = 0xFF,
                        TimingClockUsedWhenSynchronisationRequired = 0xF8,
                        StartCurrentSequence = 0xFA,
                        ContinueStoppedSequence = 0xFB,
                        StopSequence = 0xFC,
                    }

                    public enum MetaCommandType {
                        SequenceNumber = 0x00,
                        Text = 0x01,
                        Copyright = 0x02,
                        TrackName = 0x03,
                        InstrumentName = 0x04,
                        Lyric = 0x05,
                        Marker = 0x06,
                        CuePoint = 0x07,
                        EndOfTrack = 0x2F,
                        SetTempo = 0x51,
                        TimeSignature = 0x58,
                        KeySignature = 0x59,
                        SequencerSpecific = 0x7F,
                        None = 0xFF,
                    }

                    public CommandType Type;
                    public MetaCommandType MetaType;
                    public int Channel;
                    public byte[] Data;
					public string DataToString {
						get { return Encoding.ASCII.GetString(this.Data);}
					}

                    public MidiCommand(Stream s, int lastEventType) {

                        this.MetaType = MetaCommandType.None;

                        int i = s.ReadByte();

                        if (i == 0xF0) {
                            // Sysex
                            uint SkipSysEx = ReadDeltaTime(s);
                            s.Seek(SkipSysEx, SeekOrigin.Current);
                            i = s.ReadByte();
                        }
                        
                        
                        if (i < 0) throw new IOException("Out of data.");


                        if ((i & 0x80) == 0x00) {
                            i = lastEventType;
                            s.Seek(-1, SeekOrigin.Current);
                        }

                        int DataSize = 2;

                        if ((i & 0xF0) == 0xF0) {
                            this.Type = (CommandType)i;
                            this.Channel = 0;

                            switch (this.Type) {
                                case CommandType.TimingClockUsedWhenSynchronisationRequired:
                                case CommandType.StartCurrentSequence:
                                case CommandType.ContinueStoppedSequence:
                                case CommandType.StopSequence:
                                    DataSize = 0;
                                    break;
                                case CommandType.MetaEvent: {
                                        i = s.ReadByte();
                                        if (i < 0) throw new IOException("Out of data.");
                                        this.MetaType = (MetaCommandType)i;

                                        DataSize = s.ReadByte();
                                        if (DataSize < 0) throw new IOException("Out of data.");
                                    } break;
                                default:
                                    throw new InvalidDataException("Unsupported command type.");
                            }
                        } else {
                            this.Type = (CommandType)(i & 0xF0);
                            this.Channel = i & 0x0F;
                            switch (this.Type) {
                                case CommandType.PatchChange:
                                case CommandType.ChannelAfterTouch:
                                    DataSize = 1;
                                    break;
                                default:
                                    break;
                            }
                        }

                        this.Data = new byte[DataSize];
						i = s.Read(this.Data, 0, this.Data.Length);
                        //if (i != this.Data.Length) throw new IOException("Out of data.");

                    }


                }


                public uint DeltaTime;
                public MidiCommand Command;

                public TrackEvent(Stream s, TrackEvent lastEvent) {
                    this.DeltaTime = ReadDeltaTime(s);
                    this.Command = new MidiCommand(s, (int)lastEvent.Command.Type | (int)lastEvent.Command.Channel);
                }

                internal static uint ReadDeltaTime(Stream s) {
                    uint DeltaTime = 0;
                    int i = 0x80;
                    while ((i & 0x80) == 0x80) {
                        i = s.ReadByte();
                        if (i < 0) throw new IOException("Out of data.");
                        DeltaTime <<= 7;
                        DeltaTime |= ((uint)i & 0x7F);
                    }
                    return DeltaTime;
                }

				public TimeSpan ConvertToTimeSpan(uint deltaTime, ushort deltaTimeTicksPerQuarter, int? tempo) {
					if (deltaTime == 0) return TimeSpan.Zero;
					byte FrameRate = (byte)(deltaTimeTicksPerQuarter >> 8);
					byte FrameDivision = (byte)deltaTimeTicksPerQuarter;
					long TimeInTicks;
					if ((FrameRate & 0x80) != 0) {
						// 10,000,000 ticks = 1 second.
						FrameRate = (byte)(-FrameRate);
						TimeInTicks = (10000000L * (long)deltaTime) / ((long)FrameRate * (long)FrameDivision);
					} else {
						if (!tempo.HasValue) throw new ArgumentException("Cannot handle time bases that do not use SMPTE standard frame rates without a tempo.");
						// Tempo = number of microseconds per quarter note.
						TimeInTicks = (10L * (long)tempo.Value * (long)deltaTime) / (long)deltaTimeTicksPerQuarter;
					}
					return new TimeSpan(TimeInTicks);
				}


                public void Write(Stream s) {
                    uint time = this.DeltaTime;
	                uint buffer = time & 0x7F;
	                byte writeval;

                    while ((time >>= 7) != 0) {
	                    buffer <<= 8;
                        buffer |= ((time & 0x7F) | 0x80);
                    }

                    for (; ; ) {
                        writeval = (byte)(buffer & 0xFF);
                        s.WriteByte(writeval);
                        if ((buffer & 0x80) != 0) {
                            buffer >>= 8;
                        } else {
                            break;
                        }
                    }


                    switch (this.Command.Type) {
                        case MidiCommand.CommandType.MetaEvent:
                            s.WriteByte(0xFF);
                            s.WriteByte((byte)this.Command.MetaType);
                            s.WriteByte((byte)this.Command.Data.Length);
                            break;
                        default:
                            s.WriteByte((byte)((byte)this.Command.Type | (byte)this.Command.Channel));
                            break;
                    }
                    s.Write(this.Command.Data, 0, this.Command.Data.Length);

                }

            }

            public TrackEvent[] Events;

            public Track(Stream s) {
                BinaryReader BR = new BinaryReader(s, Encoding.ASCII);
                string Header = new string(BR.ReadChars(4));
                if (Header != "MTrk") throw new InvalidDataException("Invalid track header.");

                uint Size = StandardMidiFile.BigEndianReadUInt32(s);
                uint FinalPosition = (uint)(s.Position + Size);
                List<TrackEvent> TrackEvents = new List<TrackEvent>(256);
                TrackEvent TE = new TrackEvent();
                while (s.Position < FinalPosition) {
                    TE = new TrackEvent(s, TE);
                    TrackEvents.Add(TE);
                    if (TE.Command.Type == TrackEvent.MidiCommand.CommandType.MetaEvent &&
                        TE.Command.MetaType == TrackEvent.MidiCommand.MetaCommandType.EndOfTrack) break;
                }
                if (s.Position != FinalPosition) {
					//TODO: Check this oddity out.
                    s.Position = FinalPosition;
                }
                this.Events = TrackEvents.ToArray();
            }

			public override string ToString() {
				string s;
				if (this.TryGetName(out s)) return s;
				return base.ToString();
			}

   
        }

        public enum TrackFormat {
            SingleTrack,
            MultipleSychronousTracks,
            MultipleAsychronousTracks,
        }

        private TrackFormat format;
        public TrackFormat Format {
            get { return this.format; }
            set {
                if ((int)value < 0 || (int)value > 2) throw new IndexOutOfRangeException("Type must be 0, 1 or 2.");
                this.format = value; 
            }
        }

        private Track[] tracks;
        public Track[] Tracks {
            get { return this.tracks; }
            set { this.tracks = value; }
        }

        private ushort deltaTimeTicksPerQuarter;
        public ushort DeltaTimeTicksPerQuarter {
            get { return this.deltaTimeTicksPerQuarter; }
            set { this.deltaTimeTicksPerQuarter = value; }
        }

        public StandardMidiFile(Stream s) {

            BinaryReader BR = new BinaryReader(s, Encoding.ASCII);

            // Check for a MThd header.
            if (new string(BR.ReadChars(4)) != "MThd") throw new InvalidDataException("Invalid file header.");

            // Check header size.
            if (BigEndianReadUInt32(s) != 6) throw new InvalidDataException("Invalid file header size.");

            // Get MIDI format.
            this.Format = (TrackFormat)BigEndianReadUInt16(s);

            // Get the number of tracks.
            this.tracks = new Track[BigEndianReadUInt16(s)];

            // Delta
            this.DeltaTimeTicksPerQuarter = BigEndianReadUInt16(s);

            for (int i = 0; i < this.tracks.Length; ++i) {
                this.tracks[i] = new Track(s);                
            }

        }

        internal static uint BigEndianReadUInt32(Stream s) {
            byte[] b =  new byte[4];
            if (s.Read(b, 0, 4) != 4) throw new IOException("Out of data.");
            return (uint)((b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3]);
        }

        internal static ushort BigEndianReadUInt16(Stream s) {
            byte[] b = new byte[2];
            if (s.Read(b, 0, 2) != 2) throw new IOException("Out of data.");
            return (ushort)((b[0] << 8) | (b[1]));
        }

        internal static void BigEndianWriteUInt32(Stream s, uint v) {
            for (int i = 0; i < 4; i++) {
                s.WriteByte((byte)(v >> 24)); v <<= 8;
            }            
        }
        internal static void BigEndianWriteUInt16(Stream s, ushort v) {
            for (int i = 0; i < 2; i++) {
                s.WriteByte((byte)(v >> 8)); v <<= 8;
            }
        }

        public void SaveType0(Stream s) {
            BinaryWriter BW = new BinaryWriter(s, Encoding.ASCII);
            BW.Write("MThd".ToCharArray());
            BigEndianWriteUInt32(s, 6);
            BigEndianWriteUInt16(s, 0);
            BigEndianWriteUInt16(s, 1);
            BigEndianWriteUInt16(s, this.DeltaTimeTicksPerQuarter);
            BW.Write("MTrk".ToCharArray());
            BigEndianWriteUInt32(s, 0x000);
            uint StartOfTrack = (uint)s.Position;

            List<Stack<Track.TrackEvent>> TracksToMerge = new List<Stack<Track.TrackEvent>>();

            foreach (Track T in this.Tracks) {
                Track.TrackEvent[] Source = T.Events;
                Array.Reverse(Source);
                TracksToMerge.Add(new Stack<Track.TrackEvent>(Source));
            }

            while (TracksToMerge.Count != 0) {

                uint ClosestEventTime = uint.MaxValue;
                Stack<Track.TrackEvent> ClosestEvent = null;

                foreach (Stack<Track.TrackEvent> Q in TracksToMerge) {
                    Track.TrackEvent Test = Q.Peek();
                    if (Test.DeltaTime < ClosestEventTime) {
                        ClosestEventTime = Test.DeltaTime;
                        ClosestEvent = Q;
                    }
                }


                // Check if it's an end-of-track marker:
                Track.TrackEvent CheckEndOfTrack = ClosestEvent.Peek();
                if (CheckEndOfTrack.Command.Type == Track.TrackEvent.MidiCommand.CommandType.MetaEvent &&
                    CheckEndOfTrack.Command.MetaType == Track.TrackEvent.MidiCommand.MetaCommandType.EndOfTrack &&
                    TracksToMerge.Count > 1) {
                    TracksToMerge.Remove(ClosestEvent);
                    continue;

                }


                // Shunt all the other track times along;
                foreach (Stack<Track.TrackEvent> Q in TracksToMerge) {
                    if (Q != ClosestEvent) {
                        Track.TrackEvent ToAdjust = Q.Pop();
                        ToAdjust.DeltaTime -= ClosestEventTime;
                        Q.Push(ToAdjust);
                    }
                }

                // Here's our event to write:
                ClosestEvent.Pop().Write(s);
                if (ClosestEvent.Count == 0) {
                    TracksToMerge.Remove(ClosestEvent);
                }

            }

            // Write the track size;
            uint SizeOfTrack = (uint)(s.Position - StartOfTrack);
            s.Seek(-(SizeOfTrack + 4), SeekOrigin.Current);
            BigEndianWriteUInt32(s, SizeOfTrack);
            s.Seek(SizeOfTrack, SeekOrigin.Current);


        }

		/// <summary>
		/// Merge the multiple MIDI tracks into a single one.
		/// </summary>
		/// <returns>The merged track.</returns>
		public Track GetMergedTrack() {
			Track Result = new Track();
			List<Track.TrackEvent> MergedEvents = new List<Track.TrackEvent>();

			// Queue up all the events:
			Stack<Track.TrackEvent>[] TracksToMerge = new Stack<Track.TrackEvent>[this.Tracks.Length];
			for (int i = 0; i < Tracks.Length; ++i) {
				Track.TrackEvent[] Source = Tracks[i].Events;
				Array.Reverse(Source);
				TracksToMerge[i] = new Stack<Track.TrackEvent>(Source);
			}

			// Now we need to merge!
			for (; ; ) {

				// Search for the next event:
				Stack<Track.TrackEvent> TrackToMergeFrom = null;
				uint NextEventTime = uint.MaxValue;
				foreach (Stack<Track.TrackEvent> FindNextEvent in TracksToMerge) {
					if (FindNextEvent.Count > 0) {
						Track.TrackEvent CheckEvent = FindNextEvent.Peek();
						if (CheckEvent.DeltaTime < NextEventTime) {
							NextEventTime = CheckEvent.DeltaTime;
							TrackToMergeFrom = FindNextEvent;
						}
					}
				}

				// Have we finished?
				if (TrackToMergeFrom == null) break;
				foreach (Stack<Track.TrackEvent> FindNextEvent in TracksToMerge) {
					if (FindNextEvent.Count == 0) continue;
					Track.TrackEvent TE = FindNextEvent.Pop();
					if (FindNextEvent == TrackToMergeFrom) {
						if (!(TE.Command.Type == Track.TrackEvent.MidiCommand.CommandType.MetaEvent && TE.Command.MetaType == Track.TrackEvent.MidiCommand.MetaCommandType.EndOfTrack)) {
							MergedEvents.Add(TE);
						}
					} else {
						TE.DeltaTime -= NextEventTime;
						FindNextEvent.Push(TE);

					}
				}
			}
			Result.Events = MergedEvents.ToArray();
			return Result;
		}

	}
}
